import torch


def sq_euclidean_distance(x, y):
    for inp in [x, y]:
        assert len(inp.shape) == 1 or inp.shape[1] == 1
    return torch.pow(x.reshape(-1, 1) - y.reshape(1, -1), 2)


def hamming_distance(x, y):
    pass



import torch
from ..kernels.rbf import rbf, rbf_init
from ..kernels.transforms import log_exp

from gpytorch.utils.cholesky import psd_safe_cholesky
from gpytorch.distributions import MultivariateNormal


class GP(torch.nn.Module):
    def __init__(self, kernel_list, input_dim, inducing_points):
        super(GP, self).__init__()
        assert len(kernel_list) == input_dim
        assert len(inducing_points) == input_dim

        self.kernel_list = kernel_list

        self.register_buffer("zero", torch.zeros(1, dtype=torch.float))
        self.register_buffer("one", torch.ones(1, dtype=torch.float))

        # Initialize parameters
        self.raw_global_variance = torch.nn.Parameter(
            torch.tensor(0.0, dtype=torch.float)
        )
        self.raw_global_noise_variance = torch.nn.Parameter(
            torch.tensor(0.0, dtype=torch.float)
        )
        self.raw_mean = torch.nn.Parameter(torch.tensor(0.0, dtype=torch.float))
        for k_i, kernel in enumerate(kernel_list):
            if kernel.__name__ == "rbf":
                rbf_init(self, k_i)
                assert inducing_points[k_i] is None
            else:
                pass
                # TODO: Register inducing points as parameters

    def compute_kern(self, X1, X2):
        kern = (log_exp(self.raw_global_variance) * self.one).repeat(
            (X1.shape[0], X2.shape[0])
        )

        for k_i, kernel in enumerate(self.kernel_list):
            if kernel.__name__ == "rbf":
                kern.mul_(
                    kernel(
                        X1[:, k_i],
                        X2[:, k_i],
                        self.get_parameter(f"raw_lengthscale{k_i}"),
                    )
                )
        return kern

    def forward(self, X, y):
        kern = self.compute_kern(X, X)

        kdiag = kern.diagonal()
        kdiag += log_exp(self.raw_global_noise_variance)

        # dist = MultivariateNormal(self.zero.repeat(X.shape[0]), kern)
        dist = MultivariateNormal(self.raw_mean.repeat(X.shape[0]), kern)
        return -dist.log_prob(y) / X.numel()

    def predict(self, X_orig, y_orig, X_pred):
        with torch.no_grad():
            kern = self.compute_kern(X_orig, X_orig)
            kdiag = kern.diagonal()
            kdiag += log_exp(self.raw_global_noise_variance)
            l_kern = psd_safe_cholesky(kern)
            alpha = torch.cholesky_solve(y_orig.reshape(-1, 1) - self.raw_mean, l_kern)
            del kern
            k_star = self.compute_kern(X_pred, X_orig)
            k_star_star = self.compute_kern(X_pred, X_pred)
            mean = k_star @ alpha + self.raw_mean
            v = torch.cholesky_solve(k_star.T, l_kern)
            cov = k_star_star - k_star @ v
            kdiag = cov.diagonal()
            kdiag += log_exp(self.raw_global_noise_variance)

            pred_dist = MultivariateNormal(mean.ravel(), cov)
            return pred_dist



import pytest
import torch
import gpytorch
from regdata import Olympic, NonStat2D
from gpytorch.kernels import RBFKernel, ScaleKernel
from skgpytorch.models import ExactGPRegressor, ExactNSGPRegressor
from skgpytorch.metrics import mean_squared_error

from nsgptorch.models import GP
from nsgptorch.kernels import rbf


@pytest.mark.eq2d
def test_equals_2d():
    n_iters = 100

    datafunc = NonStat2D(backend="torch")
    X_train, y_train, X_test = map(lambda x: x.to(torch.float32), datafunc.get_data())
    y_test = datafunc.f(X_test[:, 0], X_test[:, 1]).to(torch.float32)

    kernel = ScaleKernel(RBFKernel(ard_num_dims=X_train.shape[1]))
    our_kernel = [rbf for _ in range(X_train.shape[1])]
    our_inducing = [None for _ in range(X_train.shape[1])]

    model = ExactGPRegressor(X_train, y_train, kernel)
    our_model = ExactNSGPRegressor(our_kernel, X_train.shape[1], our_inducing)

    for seed in range(10, 15):
        model.fit(n_iters=n_iters, random_state=seed)
        our_model.fit(X_train, y_train, n_iters=n_iters, random_state=seed)

        dist = model.predict(X_train, y_train, X_test)
        our_dist = our_model.predict(X_train, y_train, X_test)

        mse = mean_squared_error(dist, y_test, squared=False)
        our_mse = mean_squared_error(our_dist, y_test, squared=False)

        assert abs(mse - our_mse) < 0.01
        # print(seed, "our", our_mse, "their", mse, "diff", abs(mse - our_mse))



import torch
from .distance import sq_euclidean_distance
from .transforms import log_exp


def rbf(x, y, lengthscale):
    """
    RBF kernel.
    """
    lengthscale = log_exp(lengthscale)

    dist = sq_euclidean_distance(x / lengthscale, y / lengthscale)
    return dist.div_(-2).exp_()


def rbf_init(model, k_i):
    model.register_parameter(
        f"raw_lengthscale{k_i}", torch.nn.Parameter(torch.zeros(1, dtype=torch.float))
    )



import os

# Manually modify following parameters to customize the structure of your project
path = os.path.abspath(os.path.dirname(__file__)).split("/")
# print(path)
REPO_HOME_PATH = "/".join(path[:-1])
REPO_NAME = path[-1]
PACKAGE_NAME = REPO_NAME
AUTHOR = "Zeel B Patel"
AUTHOR_EMAIL = "patel_zeel@iitgn.ac.in"
description = "Non-stationary GP package written from scratch in PyTorch"
URL = "https://github.com/patel-zeel/" + REPO_NAME
LICENSE = "MIT"
LICENSE_FILE = "LICENSE"
LONG_DESCRIPTION = "file: README.md"
LONG_DESCRIPTION_CONTENT_TYPE = "text/markdown"

full_path = os.path.join(REPO_HOME_PATH, REPO_NAME)

# Write setup.cfg

with open(os.path.join(full_path, "setup.cfg"), "w") as f:
    f.write("[metadata]\n")
    f.write("name = " + PACKAGE_NAME + "\n")
    f.write("author = " + AUTHOR + "\n")
    f.write("author-email = " + AUTHOR_EMAIL + "\n")
    f.write("description = " + description + "\n")
    f.write("url = " + URL + "\n")
    f.write("license = " + LICENSE + "\n")
    f.write("long_description_content_type = " + LONG_DESCRIPTION_CONTENT_TYPE + "\n")
    f.write("long_description = " + LONG_DESCRIPTION + "\n")

# Write CI

with open(os.path.join(full_path, ".github/workflows/CI.template"), "r") as f:
    content = f.read()

with open(os.path.join(full_path, ".github/workflows/CI.yml"), "w") as f:
    content = content.replace("<reponame>", REPO_NAME)
    f.write(content)

# Write .gitignore
with open(os.path.join(full_path, ".gitignore"), "w") as f:
    f.write("__pycache__/\n")
    f.write("*.vscode\n")
    f.write("*.pyc\n")
    f.write("*.egg-info/\n")
    f.write(f"{PACKAGE_NAME}/_version.py\n")


# Write pyproject.toml
with open(os.path.join(full_path, "pyproject.toml"), "w") as f:
    f.write("[build-system]\n")
    f.write("requires = [\n")
    f.write('\t"setuptools>=50.0",\n')
    f.write('\t"setuptools_scm[toml]>=6.0",\n')
    f.write('\t"setuptools_scm_git_archive",\n')
    f.write('\t"wheel>=0.33",\n')
    f.write('\t"numpy>=1.16",\n')
    f.write('\t"cython>=0.29",\n')
    f.write("\t]\n")
    f.write("\n")
    f.write("[tool.setuptools_scm]\n")
    f.write(f'write_to = "{PACKAGE_NAME}/_version.py"')

# Write requirements.txt
with open(os.path.join(full_path, "requirements.txt"), "r") as f:
    data = f.read()

with open(os.path.join(full_path, "requirements.txt"), "w") as f:
    f.write(data)
    if "setuptools_scm[toml]" not in data:
        f.write("\nsetuptools_scm[toml]\n")
    if "setuptools_scm_git_archive" not in data:
        f.write("\nsetuptools_scm_git_archive\n")

# Initialize project folder
os.makedirs(os.path.join(full_path, PACKAGE_NAME))
with open(os.path.join(full_path, PACKAGE_NAME, "__init__.py"), "w") as f:
    f.write("from ._version import version as __version__  # noqa")

print("Successful")



from setuptools import find_packages, setup

with open('requirements.txt') as f:
    requirements = f.read().strip().split('\n')

setup(
    packages=find_packages(),
    python_requires=">=3.6",
    install_requires=requirements,
    include_package_data=True,
)